// Bindings generated by `windows-bindgen` 0.51.1

#![allow(
    non_snake_case,
    non_upper_case_globals,
    non_camel_case_types,
    dead_code,
    clippy::all
)]
#[link(name = "kernel32")]
extern "system" {
    pub fn GetUserPreferredUILanguages(
        dwflags: u32,
        pulnumlanguages: *mut u32,
        pwszlanguagesbuffer: PWSTR,
        pcchlanguagesbuffer: *mut u32,
    ) -> BOOL;
}
pub type BOOL = i32;
pub const MUI_LANGUAGE_NAME: u32 = 8u32;
pub type PWSTR = *mut u16;
pub const TRUE: BOOL = 1i32;
